#!/usr/bin/python3

import argparse
import sys
import itertools
import random

class UnionFind(object):
    def __init__(self, n):
        self.parent = list(range(n))
        self.size = [1] * n
        self.components = n
    
    def union(self, u, v):
        u = self.find(u)
        v = self.find(v)
        if u == v: return
        
        self.components -= 1

        if self.size[u] < self.size[v]:
            u, v = v, u
        self.parent[v] = u
        self.size[u] += self.size[v]

    def find(self, u):
        while u != self.parent[u]:
            self.parent[u] = self.parent[self.parent[u]]
            u = self.parent[u]
        return u


def main():
    parser = argparse.ArgumentParser(description="Generator for server connections.")
    parser.add_argument('-o', metavar='O', type=str, default="",
                        help='Options for generation')
    parser.add_argument('-n', metavar='N', type=int, default=4)
    parser.add_argument('-m', metavar='M', type=int, default=6)
    parser.add_argument('-s', metavar='S', type=int, default=None,
                        help='seed for generation (optional)')
    parser.add_argument('-t', metavar='T', type=int, default=1)
    parser.add_argument('-z', metavar='Z', type=int, default=2)
    parser.add_argument('--budget-ok', dest='budg', action='store_true')
    parser.add_argument('--no-budget-ok', dest='budg', action='store_false')
    parser.add_argument('--constraints-ok', dest='cons', action='store_true')
    parser.add_argument('--no-constraints-ok', dest='cons', action='store_false')
    parser.set_defaults(budg=True, cons=True)


    args = parser.parse_args()
    random.seed(args.s)
    # assume n and m are given as input parameters
    n = args.n
    m = args.m
    t = args.t
    z = args.z
    budget_ok = args.budg
    constraints_ok = args.cons
    k_limit = 50000

    # print("debug: n=%d, m=%d, t=%d, z=%d, bud=%s, cons=%s" % (n, m, t, z, str(budget_ok), str(constraints_ok)))

    (g_n, _), constraints, edges = gen_random(n - z, m - z + 1, constraints_ok)


    t_constraints, t_edges, n_t = iterative_trees(z, t, g_n)

    final_n = g_n + len(t_constraints)
    final_m = len(edges) + len(t_edges)
    final_edges = edges
    for u, v in t_edges:
        final_edges.add((u, v))
    final_constraints = constraints + t_constraints

    adj = [[] for _ in range(final_n)]
    for u, v in final_edges:
        adj[u].append(v)
        adj[v].append(u)
    
    final_k = min(k_limit, solve_for_k(adj, final_constraints))

    if not budget_ok:
        final_k -= 1    

    assert len(final_constraints) == final_n
    assert len(final_edges) == final_m
    print("%d %d %d" % (final_n, final_m, final_k))
    print(" ".join(str(x) for x in final_constraints))
    for u, v in final_edges:
        print("%d %d" % (u, v))

def gen_saturated_tree(n):
    edges = set()
    constraints = [0] * n
    uf = UnionFind(n)
    while uf.components > 1:
        u = random.randint(0, n - 1)
        v = random.randint(0, n - 1)
        if uf.find(u) == uf.find(v): continue
        edges.add((u, v))
        uf.union(u, v)
        constraints[u] += 1
        constraints[v] += 1
    return (constraints, edges)

def iterative_trees(n, num_trees, offset=0):
    edges = set()
    generated = 0
    constraints = []
    nt = 0
    while generated < n and nt < num_trees:
        size = random.randint(2, int(n - generated / num_trees))
        if generated + size == n - 1:
            size += 1
        n_constraints, n_edges = gen_saturated_tree(size)
        constraints += n_constraints
        for u, v in n_edges:
            edges.add((u + offset + generated, v + offset + generated))
        generated += size
        nt += 1
    return (constraints, edges, nt)

def gen_saturated_trees(n, num_trees, offset=0):
    edges = set()
    generated = 0
    constraints = [0] * n
    nt = 0
    while generated < n and nt < num_trees:
        size = random.randint(2, int(n - generated / num_trees))
        if generated + size == n - 1:
            size += 1
        uf = UnionFind(size)
        temp_offset = offset + generated
        new_edges = set()
        while uf.components > 1:
            u = random.randint(temp_offset, temp_offset + size - 1)
            v = random.randint(temp_offset, temp_offset + size - 1)
            if uf.find(u - temp_offset) == uf.find(v - temp_offset): continue
            if u > v:
                u, v = v, u
            new_edges.add((u, v))
            uf.union(u - temp_offset, v - temp_offset)
            constraints[u - offset] += 1
            constraints[v - offset] += 1
        for e in new_edges:
            edges.add(e)
        generated += size
        nt += 1
        print("generated: %d" % generated)
        
    return (constraints, edges, nt)

def gen_random(n, m, constraints_ok):
    adj = [[] for _ in range(n)]
    edges = set()
    while len(edges) < m:
        u = random.randint(0, n - 1)
        v = random.randint(0, n - 1)
        if u == v: continue
        if v < u:
            u, v = v, u
        edges.add((u, v))

    for u, v in edges:
        adj[u].append(v)
        adj[v].append(u)
    
    min_constraints = []
    for i in range(n):
        min_constraints.append(max(1, len(adj[i])))

    total = sum(min_constraints)
    min_possible_total = (n - 1) * 2
    extra_needed = max(0, min_possible_total - total)
    if not constraints_ok and extra_needed > 0:
        extra_needed -= 1

    for _ in range(extra_needed):
        min_constraints[random.randint(0, n - 1)] += 1
    
    return ((n, m), min_constraints, edges)

    # print("%d %d %d" % (n, m, k))
    # print(" ".join(str(x) for x in min_constraints))
    # print("\n".join("%d %d" % e for e in edges))
    

def find_cc_caps(adj, constraints):
    n = len(adj)
    visited = [False] * n
    ccs = []
    for i in range(n):
        if visited[i]:
            continue
        stack = [i]
        comp_free = 0
        comp_edges = 0
        comp_size = 0
        while len(stack) > 0:
            u = stack.pop()
            if visited[u]:
                continue
            visited[u] = True
            comp_size += 1
            comp_edges += len(adj[u])
            comp_free += constraints[u] - len(adj[u])
            for v in adj[u]:
                if not visited[v]:
                    stack.append(v)
        comp_edges /= 2
        ccs.append((comp_free, comp_free + 2 * (comp_edges - comp_size + 1)))
    return ccs


def solve_for_k(adj, constraints):
    # finds the minimum k that completes this case, assuming it is possible

    ccs = find_cc_caps(adj, constraints)

    if len(ccs) == 1:
        return 0

    # The following solution uses a formula to 
    # determine whether all components can be connected
    all_total = 0
    all_free = 0
    sated = 0
    full = 0

    for fre, tot in ccs:
        all_total += tot
        all_free += fre
        if fre == 0 and tot == 0:
            sated += 1
        elif fre == 0:
            full += 1
    
    x = max(len(ccs) - 1 - full - int(all_free / 2), 0)
    
    return x + (len(ccs) - 1) + full + 2 * sated



if __name__ == '__main__':
    main()
